@@ -47,7 +47,8 @@ enum _wbcErrType {
 	WBC_ERR_AUTH_ERROR,        /**< Authentication failed **/
 	WBC_ERR_UNKNOWN_USER,      /**< User account cannot be found */
 	WBC_ERR_UNKNOWN_GROUP,     /**< Group account cannot be found */
-	WBC_ERR_PWD_CHANGE_FAILED  /**< Password Change has failed */
+	WBC_ERR_PWD_CHANGE_FAILED, /**< Password Change has failed */
+	WBC_ERR_NOT_MAPPED         /**< Translation function failed on some element **/
 };
 
 typedef enum _wbcErrType wbcErr;
@@ -333,6 +334,7 @@ struct wbcChangePasswordParams {
 /* wbcAuthUserParams->flags */
 
 #define WBC_AUTH_PARAM_FLAGS_INTERACTIVE_LOGON		0x00000001
+#define WBC_AUTH_PARAM_FLAGS_FOR_NETLOGON		0x00000002
 
 /**
  * @brief Auth User Information
@@ -417,7 +419,7 @@ struct wbcLogonUserInfo {
 #define WBC_ACB_SVRTRUST			0x00000100 /* 1 Server trust account */
 #define WBC_ACB_PWNOEXP				0x00000200 /* 1 User password does not expire */
 #define WBC_ACB_AUTOLOCK			0x00000400 /* 1 Account auto locked */
-#define WBC_ACB_ENC_TXT_PWD_ALLOWED		0x00000800 /* 1 Encryped text password is allowed */
+#define WBC_ACB_ENC_TXT_PWD_ALLOWED		0x00000800 /* 1 Encrypted text password is allowed */
 #define WBC_ACB_SMARTCARD_REQUIRED		0x00001000 /* 1 Smart Card required */
 #define WBC_ACB_TRUSTED_FOR_DELEGATION		0x00002000 /* 1 Trusted for Delegation */
 #define WBC_ACB_NOT_DELEGATED			0x00004000 /* 1 Not delegated */
@@ -1097,7 +1099,7 @@ wbcErr wbcAllocateGid(gid_t *pgid);
  *
  * @return #wbcErr
  *
- * @deprecated      This method is not impemented any more and should
+ * @deprecated      This method is not implemented any more and should
  *                  be removed in the next major version change.
  **/
 wbcErr wbcSetUidMapping(uid_t uid, const struct wbcDomainSid *sid);
@@ -1110,7 +1112,7 @@ wbcErr wbcSetUidMapping(uid_t uid, const struct wbcDomainSid *sid);
  *
  * @return #wbcErr
  *
- * @deprecated      This method is not impemented any more and should
+ * @deprecated      This method is not implemented any more and should
  *                  be removed in the next major version change.
  **/
 wbcErr wbcSetGidMapping(gid_t gid, const struct wbcDomainSid *sid);
@@ -1123,7 +1125,7 @@ wbcErr wbcSetGidMapping(gid_t gid, const struct wbcDomainSid *sid);
  *
  * @return #wbcErr
  *
- * @deprecated      This method is not impemented any more and should
+ * @deprecated      This method is not implemented any more and should
  *                  be removed in the next major version change.
  **/
 wbcErr wbcRemoveUidMapping(uid_t uid, const struct wbcDomainSid *sid);
@@ -1136,7 +1138,7 @@ wbcErr wbcRemoveUidMapping(uid_t uid, const struct wbcDomainSid *sid);
  *
  * @return #wbcErr
  *
- * @deprecated      This method is not impemented any more and should
+ * @deprecated      This method is not implemented any more and should
  *                  be removed in the next major version change.
  **/
 wbcErr wbcRemoveGidMapping(gid_t gid, const struct wbcDomainSid *sid);
@@ -1148,7 +1150,7 @@ wbcErr wbcRemoveGidMapping(gid_t gid, const struct wbcDomainSid *sid);
  *
  * @return #wbcErr
  *
- * @deprecated      This method is not impemented any more and should
+ * @deprecated      This method is not implemented any more and should
  *                  be removed in the next major version change.
  **/
 wbcErr wbcSetUidHwm(uid_t uid_hwm);
@@ -1160,7 +1162,7 @@ wbcErr wbcSetUidHwm(uid_t uid_hwm);
  *
  * @return #wbcErr
  *
- * @deprecated      This method is not impemented any more and should
+ * @deprecated      This method is not implemented any more and should
  *                  be removed in the next major version change.
  **/
 wbcErr wbcSetGidHwm(gid_t gid_hwm);
