@@ -24,92 +24,37 @@
 #include "../librpc/gen_ndr/rap.h"
 #include "../lib/util/tevent_ntstatus.h"
 #include "async_smb.h"
-#include "libsmb/libsmb.h"
+#include "source3/include/client.h"
+#include "source3/libsmb/proto.h"
 #include "libsmb/clirap.h"
 #include "trans2.h"
 #include "../libcli/smb/smbXcli_base.h"
+#include "libcli/smb/reparse.h"
 #include "cli_smb2_fnum.h"
 #include "lib/util/string_wrappers.h"
 
 #include <gnutls/gnutls.h>
 #include <gnutls/crypto.h>
 
-#define PIPE_LANMAN   "\\PIPE\\LANMAN"
-
-/****************************************************************************
- Call a remote api
-****************************************************************************/
-
-bool cli_api(struct cli_state *cli,
-	     char *param, int prcnt, int mprcnt,
-	     char *data, int drcnt, int mdrcnt,
-	     char **rparam, unsigned int *rprcnt,
-	     char **rdata, unsigned int *rdrcnt)
-{
-	NTSTATUS status;
-
-	uint8_t *my_rparam, *my_rdata;
-	uint32_t num_my_rparam, num_my_rdata;
-
-	status = cli_trans(talloc_tos(), cli, SMBtrans,
-			   PIPE_LANMAN, 0, /* name, fid */
-			   0, 0,	   /* function, flags */
-			   NULL, 0, 0,	   /* setup */
-			   (uint8_t *)param, prcnt, mprcnt, /* Params, length, max */
-			   (uint8_t *)data, drcnt, mdrcnt,  /* Data, length, max */
-			   NULL,		 /* recv_flags2 */
-			   NULL, 0, NULL,	 /* rsetup */
-			   &my_rparam, 0, &num_my_rparam,
-			   &my_rdata, 0, &num_my_rdata);
-	if (!NT_STATUS_IS_OK(status)) {
-		return false;
-	}
-
-	/*
-	 * I know this memcpy massively hurts, but there are just tons
-	 * of callers of cli_api that eventually need changing to
-	 * talloc
-	 */
-
-	*rparam = (char *)smb_memdup(my_rparam, num_my_rparam);
-	if (*rparam == NULL) {
-		goto fail;
-	}
-	*rprcnt = num_my_rparam;
-	TALLOC_FREE(my_rparam);
-
-	*rdata = (char *)smb_memdup(my_rdata, num_my_rdata);
-	if (*rdata == NULL) {
-		goto fail;
-	}
-	*rdrcnt = num_my_rdata;
-	TALLOC_FREE(my_rdata);
-
-	return true;
-fail:
-	TALLOC_FREE(my_rdata);
-	TALLOC_FREE(my_rparam);
-	*rparam = NULL;
-	*rprcnt = 0;
-	*rdata = NULL;
-	*rdrcnt = 0;
-	return false;
-}
-
 /****************************************************************************
  Call a NetShareEnum - try and browse available connections on a host.
 ****************************************************************************/
 
-int cli_RNetShareEnum(struct cli_state *cli, void (*fn)(const char *, uint32_t, const char *, void *), void *state)
+NTSTATUS cli_RNetShareEnum(
+	struct cli_state *cli,
+	void (*fn)(const char *, uint32_t, const char *, void *),
+	void *state)
 {
-	char *rparam = NULL;
-	char *rdata = NULL;
-	char *p;
+	uint8_t *rparam = NULL;
+	uint8_t *rdata = NULL;
+	char *rdata_end = NULL;
+	char *p = NULL;
 	unsigned int rdrcnt,rprcnt;
 	char param[1024];
 	int count = -1;
-	bool ok;
+	int i, converter;
 	int res;
+	NTSTATUS status;
 
 	/* now send a SMBtrans command with api RNetShareEnum */
 	p = param;
@@ -127,86 +72,102 @@ int cli_RNetShareEnum(struct cli_state *cli, void (*fn)(const char *, uint32_t,
 	SSVAL(p,2,0xFFE0);
 	p += 4;
 
-	ok = cli_api(
-		cli,
-		param, PTR_DIFF(p,param), 1024,  /* Param, length, maxlen */
-		NULL, 0, 0xFFE0,            /* data, length, maxlen - Win2k needs a small buffer here too ! */
-		&rparam, &rprcnt,                /* return params, length */
-		&rdata, &rdrcnt);                /* return data, length */
-	if (!ok) {
+	status = cli_trans(talloc_tos(),       /* mem_ctx */
+			   cli,		       /* cli */
+			   SMBtrans,	       /* cmd */
+			   "\\PIPE\\LANMAN",	       /* name */
+			   0,		       /* fid */
+			   0,		       /* function */
+			   0,		       /* flags */
+			   NULL,	       /* setup */
+			   0,		       /* num_setup */
+			   0,		       /* max_setup */
+			   (uint8_t *)param,   /* param */
+			   PTR_DIFF(p, param), /* num_param */
+			   1024,	       /* max_param */
+			   NULL,	       /* data */
+			   0,		       /* num_data */
+			   0xFFE0,	       /* max_data, for W2K */
+			   NULL,	       /* recv_flags2 */
+			   NULL,	       /* rsetup */
+			   0,		       /* min_rsetup */
+			   NULL,	       /* num_rsetup */
+			   &rparam,	       /* rparam */
+			   6,		       /* min_rparam */
+			   &rprcnt,	       /* num_rparam */
+			   &rdata,	       /* rdata */
+			   0,		       /* min_rdata */
+			   &rdrcnt);	       /* num_rdata */
+	if (!NT_STATUS_IS_OK(status)) {
 		DEBUG(4,("NetShareEnum failed\n"));
 		goto done;
 	}
 
-	if (rprcnt < 6) {
-		DBG_ERR("Got invalid result: rprcnt=%u\n", rprcnt);
+	res = PULL_LE_U16(rparam, 0);
+
+	if (!(res == 0 || res == ERRmoredata)) {
+		DEBUG(4,("NetShareEnum res=%d\n", res));
+		status = werror_to_ntstatus(W_ERROR(res));
 		goto done;
 	}
 
-	res = rparam? SVAL(rparam,0) : -1;
+	converter = SVAL(rparam,2);
+	rdata_end = (char *)rdata + rdrcnt;
 
-	if (res == 0 || res == ERRmoredata) {
-		int converter=SVAL(rparam,2);
-		int i;
-		char *rdata_end = rdata + rdrcnt;
+	count=SVAL(rparam,4);
+	p = (char *)rdata;
 
-		count=SVAL(rparam,4);
-		p = rdata;
-
-		for (i=0;i<count;i++,p+=20) {
-			char *sname;
-			int type;
-			int comment_offset;
-			const char *cmnt;
-			const char *p1;
-			char *s1, *s2;
-			size_t len;
-			TALLOC_CTX *frame = talloc_stackframe();
-
-			if (p + 20 > rdata_end) {
-				TALLOC_FREE(frame);
-				break;
-			}
-
-			sname = p;
-			type = SVAL(p,14);
-			comment_offset = (IVAL(p,16) & 0xFFFF) - converter;
-			if (comment_offset < 0 ||
-			    comment_offset > (int)rdrcnt) {
-				TALLOC_FREE(frame);
-				break;
-			}
-			cmnt = comment_offset?(rdata+comment_offset):"";
-
-			/* Work out the comment length. */
-			for (p1 = cmnt, len = 0; *p1 &&
-				     p1 < rdata_end; len++)
-				p1++;
-			if (!*p1) {
-				len++;
-			}
-			pull_string_talloc(frame,rdata,0,
-					   &s1,sname,14,STR_ASCII);
-			pull_string_talloc(frame,rdata,0,
-					   &s2,cmnt,len,STR_ASCII);
-			if (!s1 || !s2) {
-				TALLOC_FREE(frame);
-				continue;
-			}
-
-			fn(s1, type, s2, state);
+	for (i=0;i<count;i++,p+=20) {
+		char *sname;
+		int type;
+		int comment_offset;
+		const char *cmnt;
+		const char *p1;
+		char *s1, *s2;
+		size_t len;
+		TALLOC_CTX *frame = talloc_stackframe();
 
+		if (p + 20 > rdata_end) {
 			TALLOC_FREE(frame);
+			break;
 		}
-	} else {
-			DEBUG(4,("NetShareEnum res=%d\n", res));
+
+		sname = p;
+		type = SVAL(p,14);
+		comment_offset = (IVAL(p,16) & 0xFFFF) - converter;
+		if (comment_offset < 0 ||
+		    comment_offset > (int)rdrcnt) {
+			TALLOC_FREE(frame);
+			break;
+		}
+		cmnt = comment_offset ? ((char *)rdata + comment_offset) : "";
+
+		/* Work out the comment length. */
+		for (p1 = cmnt, len = 0; *p1 &&
+			     p1 < rdata_end; len++)
+			p1++;
+		if (!*p1) {
+			len++;
+		}
+		pull_string_talloc(frame,rdata,0,
+				   &s1,sname,14,STR_ASCII);
+		pull_string_talloc(frame,rdata,0,
+				   &s2,cmnt,len,STR_ASCII);
+		if (!s1 || !s2) {
+			TALLOC_FREE(frame);
+			continue;
+		}
+
+		fn(s1, type, s2, state);
+
+		TALLOC_FREE(frame);
 	}
 
 done:
-	SAFE_FREE(rparam);
-	SAFE_FREE(rdata);
+	TALLOC_FREE(rparam);
+	TALLOC_FREE(rdata);
 
-	return count;
+	return status;
 }
 
 /****************************************************************************
@@ -217,14 +178,17 @@ done:
  the comment and a state pointer.
 ****************************************************************************/
 
-bool cli_NetServerEnum(struct cli_state *cli, char *workgroup, uint32_t stype,
-		       void (*fn)(const char *, uint32_t, const char *, void *),
-		       void *state)
+NTSTATUS cli_NetServerEnum(
+	struct cli_state *cli,
+	char *workgroup,
+	uint32_t stype,
+	void (*fn)(const char *, uint32_t, const char *, void *),
+	void *state)
 {
-	char *rparam = NULL;
-	char *rdata = NULL;
+	uint8_t *rparam = NULL;
+	uint8_t *rdata = NULL;
 	char *rdata_end = NULL;
-	unsigned int rdrcnt,rprcnt;
+	uint32_t rdrcnt, rprcnt;
 	char *p;
 	char param[1024];
 	int uLevel = 1;
@@ -234,8 +198,7 @@ bool cli_NetServerEnum(struct cli_state *cli, char *workgroup, uint32_t stype,
 	int total_cnt = 0;
 	int return_cnt = 0;
 	int res;
-
-	errno = 0; /* reset */
+	NTSTATUS status;
 
 	/*
 	 * This may take more than one transaction, so we should loop until
@@ -274,7 +237,7 @@ bool cli_NetServerEnum(struct cli_state *cli, char *workgroup, uint32_t stype,
 
 		if (len == 0) {
 			SAFE_FREE(last_entry);
-			return false;
+			return NT_STATUS_INTERNAL_ERROR;
 		}
 		p += len;
 
@@ -286,7 +249,7 @@ bool cli_NetServerEnum(struct cli_state *cli, char *workgroup, uint32_t stype,
 
 			if (len == 0) {
 				SAFE_FREE(last_entry);
-				return false;
+				return NT_STATUS_INTERNAL_ERROR;
 			}
 			p += len;
 		}
@@ -294,29 +257,43 @@ bool cli_NetServerEnum(struct cli_state *cli, char *workgroup, uint32_t stype,
 		/* Next time through we need to use the continue api */
 		func = RAP_NetServerEnum3;
 
-		if (!cli_api(cli,
-			param, PTR_DIFF(p,param), 8, /* params, length, max */
-			NULL, 0, CLI_BUFFER_SIZE, /* data, length, max */
-		            &rparam, &rprcnt, /* return params, return size */
-		            &rdata, &rdrcnt)) { /* return data, return size */
-
+		status = cli_trans(talloc_tos(),       /* mem_ctx */
+				   cli,		       /* cli */
+				   SMBtrans,	       /* cmd */
+				   "\\PIPE\\LANMAN",	       /* name */
+				   0,		       /* fid */
+				   0,		       /* function */
+				   0,		       /* flags */
+				   NULL,	       /* setup */
+				   0,		       /* num_setup */
+				   0,		       /* max_setup */
+				   (uint8_t *)param,   /* param */
+				   PTR_DIFF(p, param), /* num_param */
+				   8,		       /* max_param */
+				   NULL,	       /* data */
+				   0,		       /* num_data */
+				   CLI_BUFFER_SIZE,    /* max_data */
+				   NULL,	       /* recv_flags2 */
+				   NULL,	       /* rsetup */
+				   0,		       /* min_rsetup */
+				   NULL,	       /* num_rsetup */
+				   &rparam,	       /* rparam */
+				   6,		       /* min_rparam */
+				   &rprcnt,	       /* num_rparam */
+				   &rdata,	       /* rdata */
+				   0,		       /* min_rdata */
+				   &rdrcnt);	       /* num_rdata */
+		if (!NT_STATUS_IS_OK(status)) {
 			/* break out of the loop on error */
 		        res = -1;
 		        break;
 		}
 
-		rdata_end = rdata + rdrcnt;
+		rdata_end = (char *)rdata + rdrcnt;
 
-		if (rprcnt < 6) {
-			DBG_ERR("Got invalid result: rprcnt=%u\n", rprcnt);
-			res = -1;
-			break;
-		}
+		res = PULL_LE_U16(rparam, 0);
 
-		res = rparam ? SVAL(rparam,0) : -1;
-
-		if (res == 0 || res == ERRmoredata ||
-                    (res != -1 && cli_errno(cli) == 0)) {
+		if (res == 0 || res == ERRmoredata) {
 			char *sname = NULL;
 			int i, count;
 			int converter=SVAL(rparam,2);
@@ -335,7 +312,7 @@ bool cli_NetServerEnum(struct cli_state *cli, char *workgroup, uint32_t stype,
 
 			/* Keep track of how many we have read */
 			return_cnt += count;
-			p = rdata;
+			p = (char *)rdata;
 
 			/* The last name in the previous NetServerEnum reply is
 			 * sent back to server in the NetServerEnum3 request
@@ -353,7 +330,7 @@ bool cli_NetServerEnum(struct cli_state *cli, char *workgroup, uint32_t stype,
 				(strncmp(last_entry, p, 16) == 0)) {
 			    count -= 1; /* Skip this entry */
 			    return_cnt = -1; /* Not part of total, so don't count. */
-			    p = rdata + 26; /* Skip the whole record */
+			    p = (char *)rdata + 26; /* Skip the whole record */
 			}
 
 			for (i = 0; i < count; i++, p += 26) {
@@ -371,7 +348,9 @@ bool cli_NetServerEnum(struct cli_state *cli, char *workgroup, uint32_t stype,
 
 				sname = p;
 				comment_offset = (IVAL(p,22) & 0xFFFF)-converter;
-				cmnt = comment_offset?(rdata+comment_offset):"";
+				cmnt = comment_offset ? ((char *)rdata +
+							 comment_offset)
+						      : "";
 
 				if (comment_offset < 0 || comment_offset >= (int)rdrcnt) {
 					TALLOC_FREE(frame);
@@ -414,61 +393,53 @@ bool cli_NetServerEnum(struct cli_state *cli, char *workgroup, uint32_t stype,
 
 			/* If we have more data, but no last entry then error out */
 			if (!last_entry && (res == ERRmoredata)) {
-			        errno = EINVAL;
 			        res = 0;
 			}
 
 		}
 
-		SAFE_FREE(rparam);
-		SAFE_FREE(rdata);
+		TALLOC_FREE(rparam);
+		TALLOC_FREE(rdata);
 	} while ((res == ERRmoredata) && (total_cnt > return_cnt));
 
-	SAFE_FREE(rparam);
-	SAFE_FREE(rdata);
+	TALLOC_FREE(rparam);
+	TALLOC_FREE(rdata);
 	SAFE_FREE(last_entry);
 
-	if (res == -1) {
-		errno = cli_errno(cli);
-	} else {
-		if (!return_cnt) {
-			/* this is a very special case, when the domain master for the
-			   work group isn't part of the work group itself, there is something
-			   wild going on */
-			errno = ENOENT;
-		}
-	    }
-
-	return(return_cnt > 0);
+	if (return_cnt == 0) {
+		return NT_STATUS_NO_MORE_ENTRIES;
+	}
+	return NT_STATUS_OK;
 }
 
 /****************************************************************************
  Send a SamOEMChangePassword command.
 ****************************************************************************/
 
-bool cli_oem_change_password(struct cli_state *cli, const char *user, const char *new_password,
-                             const char *old_password)
+NTSTATUS cli_oem_change_password(struct cli_state *cli,
+				 const char *user,
+				 const char *new_password,
+				 const char *old_password)
 {
 	char param[1024];
-	unsigned char data[532];
+	uint8_t data[532];
 	char *p = param;
 	unsigned char old_pw_hash[16];
 	unsigned char new_pw_hash[16];
-	unsigned int data_len;
 	unsigned int param_len = 0;
-	char *rparam = NULL;
-	char *rdata = NULL;
-	unsigned int rprcnt, rdrcnt;
+	uint8_t *rparam = NULL;
+	uint32_t rprcnt;
 	gnutls_cipher_hd_t cipher_hnd = NULL;
 	gnutls_datum_t old_pw_key = {
 		.data = old_pw_hash,
 		.size = sizeof(old_pw_hash),
 	};
-	int rc;
+	int rc, res;
+	NTSTATUS status;
 
 	if (strlen(user) >= sizeof(fstring)-1) {
-		DEBUG(0,("cli_oem_change_password: user name %s is too long.\n", user));
-		return False;
+		DBG_ERR("user name %s is too long.\n", user);
+		return NT_STATUS_NAME_TOO_LONG;
 	}
 
 	SSVAL(p,0,214); /* SamOEMChangePassword command. */
@@ -503,14 +474,18 @@ bool cli_oem_change_password(struct cli_state *cli, const char *user, const char
 	if (rc < 0) {
 		DBG_ERR("gnutls_cipher_init failed: %s\n",
 			gnutls_strerror(rc));
-		return false;
+		status = gnutls_error_to_ntstatus(
+			rc, NT_STATUS_CRYPTO_SYSTEM_INVALID);
+		return status;
 	}
 	rc = gnutls_cipher_encrypt(cipher_hnd,
 			      data,
 			      516);
 	gnutls_cipher_deinit(cipher_hnd);
 	if (rc < 0) {
-		return false;
+		status = gnutls_error_to_ntstatus(
+			rc, NT_STATUS_CRYPTO_SYSTEM_INVALID);
+		return status;
 	}
 
 	/*
@@ -521,165 +496,46 @@ bool cli_oem_change_password(struct cli_state *cli, const char *user, const char
 	rc = E_old_pw_hash( new_pw_hash, old_pw_hash, (uchar *)&data[516]);
 	if (rc != 0) {
 		DBG_ERR("E_old_pw_hash failed: %s\n", gnutls_strerror(rc));
-		return false;
-	}
-
-	data_len = 532;
-
-	if (!cli_api(cli,
-		     param, param_len, 4,		/* param, length, max */
-		     (char *)data, data_len, 0,		/* data, length, max */
-		     &rparam, &rprcnt,
-		     &rdata, &rdrcnt)) {
-		DEBUG(0,("cli_oem_change_password: Failed to send password change for user %s\n",
-			user ));
-		return False;
-	}
-
-	if (rdrcnt < 2) {
-		cli->rap_error = ERRbadformat;
-		goto done;
-	}
-
-	if (rparam) {
-		cli->rap_error = SVAL(rparam,0);
-	}
-
-done:
-	SAFE_FREE(rparam);
-	SAFE_FREE(rdata);
-
-	return (cli->rap_error == 0);
-}
-
-/****************************************************************************
- Send a qpathinfo call.
-****************************************************************************/
-
-struct cli_qpathinfo1_state {
-	struct cli_state *cli;
-	uint32_t num_data;
-	uint8_t *data;
-};
-
-static void cli_qpathinfo1_done(struct tevent_req *subreq);
-
-struct tevent_req *cli_qpathinfo1_send(TALLOC_CTX *mem_ctx,
-				       struct tevent_context *ev,
-				       struct cli_state *cli,
-				       const char *fname)
-{
-	struct tevent_req *req = NULL, *subreq = NULL;
-	struct cli_qpathinfo1_state *state = NULL;
-
-	req = tevent_req_create(mem_ctx, &state, struct cli_qpathinfo1_state);
-	if (req == NULL) {
-		return NULL;
-	}
-	state->cli = cli;
-	subreq = cli_qpathinfo_send(state, ev, cli, fname, SMB_INFO_STANDARD,
-				    22, CLI_BUFFER_SIZE);
-	if (tevent_req_nomem(subreq, req)) {
-		return tevent_req_post(req, ev);
-	}
-	tevent_req_set_callback(subreq, cli_qpathinfo1_done, req);
-	return req;
-}
-
-static void cli_qpathinfo1_done(struct tevent_req *subreq)
-{
-	struct tevent_req *req = tevent_req_callback_data(
-		subreq, struct tevent_req);
-	struct cli_qpathinfo1_state *state = tevent_req_data(
-		req, struct cli_qpathinfo1_state);
-	NTSTATUS status;
-
-	status = cli_qpathinfo_recv(subreq, state, &state->data,
-				    &state->num_data);
-	TALLOC_FREE(subreq);
-	if (!NT_STATUS_IS_OK(status)) {
-		tevent_req_nterror(req, status);
-		return;
-	}
-	tevent_req_done(req);
-}
-
-NTSTATUS cli_qpathinfo1_recv(struct tevent_req *req,
-			     time_t *change_time,
-			     time_t *access_time,
-			     time_t *write_time,
-			     off_t *size,
-			     uint32_t *pattr)
-{
-	struct cli_qpathinfo1_state *state = tevent_req_data(
-		req, struct cli_qpathinfo1_state);
-	NTSTATUS status;
-
-	time_t (*date_fn)(const void *buf, int serverzone);
-
-	if (tevent_req_is_nterror(req, &status)) {
+		status = gnutls_error_to_ntstatus(
+			rc, NT_STATUS_CRYPTO_SYSTEM_INVALID);
 		return status;
 	}
 
-	if (state->cli->win95) {
-		date_fn = make_unix_date;
-	} else {
-		date_fn = make_unix_date2;
+	status = cli_trans(talloc_tos(),     /* mem_ctx */
+			   cli,		     /* cli */
+			   SMBtrans,	     /* cmd */
+			   "\\PIPE\\LANMAN", /* name */
+			   0,		     /* fid */
+			   0,		     /* function */
+			   0,		     /* flags */
+			   NULL,	     /* setup */
+			   0,		     /* num_setup */
+			   0,		     /* max_setup */
+			   (uint8_t *)param, /* param */
+			   param_len,	     /* num_param */
+			   4,		     /* max_param */
+			   data,	     /* data */
+			   sizeof(data),     /* num_data */
+			   0,		     /* max_data */
+			   NULL,	     /* recv_flags2 */
+			   NULL,	     /* rsetup */
+			   0,		     /* min_rsetup */
+			   NULL,	     /* num_rsetup */
+			   &rparam,	     /* rparam */
+			   2,		     /* min_rparam */
+			   &rprcnt,	     /* num_rparam */
+			   NULL,	     /* rdata */
+			   0,		     /* min_rdata */
+			   NULL);	     /* num_rdata */
+	if (!NT_STATUS_IS_OK(status)) {
+		return status;
 	}
+	res = PULL_LE_U16(rparam, 0);
 
-	if (change_time) {
-		*change_time = date_fn(state->data+0, smb1cli_conn_server_time_zone(state->cli->conn));
-	}
-	if (access_time) {
-		*access_time = date_fn(state->data+4, smb1cli_conn_server_time_zone(state->cli->conn));
-	}
-	if (write_time) {
-		*write_time = date_fn(state->data+8, smb1cli_conn_server_time_zone(state->cli->conn));
-	}
-	if (size) {
-		*size = IVAL(state->data, 12);
-	}
-	if (pattr) {
-		*pattr = SVAL(state->data, l1_attrFile);
-	}
-	return NT_STATUS_OK;
-}
+	status = werror_to_ntstatus(W_ERROR(res));
 
-NTSTATUS cli_qpathinfo1(struct cli_state *cli,
-			const char *fname,
-			time_t *change_time,
-			time_t *access_time,
-			time_t *write_time,
-			off_t *size,
-			uint32_t *pattr)
-{
-	TALLOC_CTX *frame = talloc_stackframe();
-	struct tevent_context *ev;
-	struct tevent_req *req;
-	NTSTATUS status = NT_STATUS_NO_MEMORY;
+	TALLOC_FREE(rparam);
 
-	if (smbXcli_conn_has_async_calls(cli->conn)) {
-		/*
-		 * Can't use sync call while an async call is in flight
-		 */
-		status = NT_STATUS_INVALID_PARAMETER;
-		goto fail;
-	}
-	ev = samba_tevent_context_init(frame);
-	if (ev == NULL) {
-		goto fail;
-	}
-	req = cli_qpathinfo1_send(frame, ev, cli, fname);
-	if (req == NULL) {
-		goto fail;
-	}
-	if (!tevent_req_poll_ntstatus(req, ev, &status)) {
-		goto fail;
-	}
-	status = cli_qpathinfo1_recv(req, change_time, access_time,
-				     write_time, size, pattr);
- fail:
-	TALLOC_FREE(frame);
 	return status;
 }
 
@@ -755,10 +611,10 @@ NTSTATUS cli_setpathinfo_ext(struct cli_state *cli, const char *fname,
 		 * the correct info type and level.
 		 */
 		return cli_smb2_setpathinfo(cli,
-				fname,
-				1, /* SMB2_SETINFO_FILE */
-				SMB_FILE_BASIC_INFORMATION - 1000,
-				&in_data);
+					    fname,
+					    SMB2_0_INFO_FILE,
+					    FSCC_FILE_BASIC_INFORMATION,
+					    &in_data);
 	}
 
 	return cli_setpathinfo(
@@ -811,9 +667,9 @@ struct tevent_req *cli_setfileinfo_ext_send(
 			cli,
 			fnum,
 			SMB2_0_INFO_FILE,
-			SMB_FILE_BASIC_INFORMATION - 1000,
+			FSCC_FILE_BASIC_INFORMATION,
 			&state->in_data,
-			0);	/* in_additional_info */
+			0); /* in_additional_info */
 		if (tevent_req_nomem(subreq, req)) {
 			return tevent_req_post(req, ev);
 		}
@@ -908,11 +764,22 @@ NTSTATUS cli_setfileinfo_ext(
 ****************************************************************************/
 
 struct cli_qpathinfo2_state {
-	uint32_t num_data;
-	uint8_t *data;
+	struct tevent_context *ev;
+	struct cli_state *cli;
+	const char *fname;
+	struct timespec create_time;
+	struct timespec access_time;
+	struct timespec write_time;
+	struct timespec change_time;
+	off_t size;
+	uint32_t attr;
+	SMB_INO_T ino;
+	mode_t mode;
 };
 
+static void cli_qpathinfo2_done2(struct tevent_req *subreq);
 static void cli_qpathinfo2_done(struct tevent_req *subreq);
+static void cli_qpathinfo2_got_reparse(struct tevent_req *subreq);
 
 struct tevent_req *cli_qpathinfo2_send(TALLOC_CTX *mem_ctx,
 				       struct tevent_context *ev,
@@ -926,6 +793,26 @@ struct tevent_req *cli_qpathinfo2_send(TALLOC_CTX *mem_ctx,
 	if (req == NULL) {
 		return NULL;
 	}
+	state->ev = ev;
+	state->cli = cli;
+	state->fname = fname;
+
+	state->mode = S_IFREG;
+
+	if (smbXcli_conn_protocol(cli->conn) >= PROTOCOL_SMB2_02) {
+		subreq = cli_smb2_qpathinfo_send(state,
+						 ev,
+						 cli,
+						 fname,
+						 FSCC_FILE_ALL_INFORMATION,
+						 0x60,
+						 UINT16_MAX);
+		if (tevent_req_nomem(subreq, req)) {
+			return tevent_req_post(req, ev);
+		}
+		tevent_req_set_callback(subreq, cli_qpathinfo2_done2, req);
+		return req;
+	}
 	subreq = cli_qpathinfo_send(state, ev, cli, fname,
 				    SMB_QUERY_FILE_ALL_INFO,
 				    68, CLI_BUFFER_SIZE);
@@ -936,21 +823,147 @@ struct tevent_req *cli_qpathinfo2_send(TALLOC_CTX *mem_ctx,
 	return req;
 }
 
+static void cli_qpathinfo2_done2(struct tevent_req *subreq)
+{
+	struct tevent_req *req =
+		tevent_req_callback_data(subreq, struct tevent_req);
+	struct cli_qpathinfo2_state *state =
+		tevent_req_data(req, struct cli_qpathinfo2_state);
+	uint8_t *rdata = NULL;
+	uint32_t num_rdata;
+	NTSTATUS status;
+
+	status = cli_smb2_qpathinfo_recv(subreq, state, &rdata, &num_rdata);
+	TALLOC_FREE(subreq);
+	if (tevent_req_nterror(req, status)) {
+		return;
+	}
+	state->create_time = interpret_long_date(BVAL(rdata, 0x0));
+	state->access_time = interpret_long_date(BVAL(rdata, 0x8));
+	state->write_time = interpret_long_date(BVAL(rdata, 0x10));
+	state->change_time = interpret_long_date(BVAL(rdata, 0x18));
+	state->attr = PULL_LE_U32(rdata, 0x20);
+	state->size = PULL_LE_U64(rdata, 0x30);
+	state->ino = PULL_LE_U64(rdata, 0x40);
+
+	if (state->attr & FILE_ATTRIBUTE_REPARSE_POINT) {
+		subreq = cli_get_reparse_data_send(state,
+						   state->ev,
+						   state->cli,
+						   state->fname);
+		if (tevent_req_nomem(subreq, req)) {
+			return;
+		}
+		tevent_req_set_callback(subreq,
+					cli_qpathinfo2_got_reparse,
+					req);
+		return;
+	}
+
+	tevent_req_done(req);
+}
+
 static void cli_qpathinfo2_done(struct tevent_req *subreq)
 {
 	struct tevent_req *req = tevent_req_callback_data(
 		subreq, struct tevent_req);
 	struct cli_qpathinfo2_state *state = tevent_req_data(
 		req, struct cli_qpathinfo2_state);
+	uint8_t *data = NULL;
+	uint32_t num_data;
 	NTSTATUS status;
 
-	status = cli_qpathinfo_recv(subreq, state, &state->data,
-				    &state->num_data);
+	status = cli_qpathinfo_recv(subreq, state, &data, &num_data);
 	TALLOC_FREE(subreq);
-	if (!NT_STATUS_IS_OK(status)) {
-		tevent_req_nterror(req, status);
+	if (tevent_req_nterror(req, status)) {
 		return;
 	}
+
+	state->create_time = interpret_long_date(BVAL(data, 0));
+	state->access_time = interpret_long_date(BVAL(data, 8));
+	state->write_time = interpret_long_date(BVAL(data, 16));
+	state->change_time = interpret_long_date(BVAL(data, 24));
+	state->attr = PULL_LE_U32(data, 32);
+	state->size = PULL_LE_U64(data, 48);
+
+	/*
+	 * SMB1 qpathinfo2 uses SMB_QUERY_FILE_ALL_INFO which doesn't
+	 * return an inode number (fileid).  We can't change this to
+	 * one of the FILE_ID info levels as only Win2003 and above
+	 * support these [MS-SMB: 2.2.2.3.1] and the SMB1 code needs
+	 * to support older servers.
+	 */
+	state->ino = 0;
+
+	TALLOC_FREE(data);
+
+	if (state->attr & FILE_ATTRIBUTE_REPARSE_POINT) {
+		subreq = cli_get_reparse_data_send(state,
+						   state->ev,
+						   state->cli,
+						   state->fname);
+		if (tevent_req_nomem(subreq, req)) {
+			return;
+		}
+		tevent_req_set_callback(subreq,
+					cli_qpathinfo2_got_reparse,
+					req);
+		return;
+	}
+
+	tevent_req_done(req);
+}
+
+static void cli_qpathinfo2_got_reparse(struct tevent_req *subreq)
+{
+	struct tevent_req *req =
+		tevent_req_callback_data(subreq, struct tevent_req);
+	struct cli_qpathinfo2_state *state =
+		tevent_req_data(req, struct cli_qpathinfo2_state);
+	uint8_t *data = NULL;
+	uint32_t num_data;
+	struct reparse_data_buffer reparse = {
+		.tag = 0,
+	};
+	NTSTATUS status;
+
+	status = cli_get_reparse_data_recv(subreq, state, &data, &num_data);
+	TALLOC_FREE(subreq);
+	if (tevent_req_nterror(req, status)) {
+		return;
+	}
+
+	status = reparse_data_buffer_parse(state, &reparse, data, num_data);
+	if (!NT_STATUS_IS_OK(status)) {
+		DBG_DEBUG("Ignoring unknown reparse data\n");
+		goto done;
+	}
+
+	switch (reparse.tag) {
+	case IO_REPARSE_TAG_SYMLINK:
+		state->mode = S_IFLNK;
+		break;
+	case IO_REPARSE_TAG_NFS:
+		switch (reparse.parsed.nfs.type) {
+		case NFS_SPECFILE_LNK:
+			state->mode = S_IFLNK;
+			break;
+		case NFS_SPECFILE_CHR:
+			state->mode = S_IFCHR;
+			break;
+		case NFS_SPECFILE_BLK:
+			state->mode = S_IFBLK;
+			break;
+		case NFS_SPECFILE_FIFO:
+			state->mode = S_IFIFO;
+			break;
+		case NFS_SPECFILE_SOCK:
+			state->mode = S_IFSOCK;
+			break;
+		}
+		break;
+	}
+done:
 	tevent_req_done(req);
 }
 
@@ -959,8 +972,10 @@ NTSTATUS cli_qpathinfo2_recv(struct tevent_req *req,
 			     struct timespec *access_time,
 			     struct timespec *write_time,
 			     struct timespec *change_time,
-			     off_t *size, uint32_t *pattr,
-			     SMB_INO_T *ino)
+			     off_t *size,
+			     uint32_t *pattr,
+			     SMB_INO_T *ino,
+			     mode_t *mode)
 {
 	struct cli_qpathinfo2_state *state = tevent_req_data(
 		req, struct cli_qpathinfo2_state);
@@ -971,65 +986,48 @@ NTSTATUS cli_qpathinfo2_recv(struct tevent_req *req,
 	}
 
 	if (create_time) {
-                *create_time = interpret_long_date((char *)state->data+0);
+		*create_time = state->create_time;
 	}
 	if (access_time) {
-		*access_time = interpret_long_date((char *)state->data+8);
+		*access_time = state->access_time;
 	}
 	if (write_time) {
-		*write_time = interpret_long_date((char *)state->data+16);
+		*write_time = state->write_time;
 	}
 	if (change_time) {
-		*change_time = interpret_long_date((char *)state->data+24);
+		*change_time = state->change_time;
 	}
 	if (pattr) {
-		/* SMB_QUERY_FILE_ALL_INFO returns 32-bit attributes. */
-		*pattr = IVAL(state->data, 32);
+		*pattr = state->attr;
 	}
 	if (size) {
-                *size = IVAL2_TO_SMB_BIG_UINT(state->data,48);
+		*size = state->size;
 	}
 	if (ino) {
-		/*
-		 * SMB1 qpathinfo2 uses SMB_QUERY_FILE_ALL_INFO
-		 * which doesn't return an inode number (fileid).
-		 * We can't change this to one of the FILE_ID
-		 * info levels as only Win2003 and above support
-		 * these [MS-SMB: 2.2.2.3.1] and the SMB1 code
-		 * needs to support older servers.
-		 */
-		*ino = 0;
+		*ino = state->ino;
+	}
+	if (mode != NULL) {
+		*mode = state->mode;
 	}
 	return NT_STATUS_OK;
 }
 
-NTSTATUS cli_qpathinfo2(struct cli_state *cli, const char *fname,
+NTSTATUS cli_qpathinfo2(struct cli_state *cli,
+			const char *fname,
 			struct timespec *create_time,
 			struct timespec *access_time,
 			struct timespec *write_time,
 			struct timespec *change_time,
-			off_t *size, uint32_t *pattr,
-			SMB_INO_T *ino)
+			off_t *size,
+			uint32_t *pattr,
+			SMB_INO_T *ino,
+			mode_t *mode)
 {
-	TALLOC_CTX *frame = NULL;
-	struct tevent_context *ev;
-	struct tevent_req *req;
+	TALLOC_CTX *frame = talloc_stackframe();
+	struct tevent_context *ev = NULL;
+	struct tevent_req *req = NULL;
 	NTSTATUS status = NT_STATUS_NO_MEMORY;
 
-	if (smbXcli_conn_protocol(cli->conn) >= PROTOCOL_SMB2_02) {
-		return cli_smb2_qpathinfo2(cli,
-					fname,
-					create_time,
-					access_time,
-					write_time,
-					change_time,
-					size,
-					pattr,
-					ino);
-	}
-
-	frame = talloc_stackframe();
-
 	if (smbXcli_conn_has_async_calls(cli->conn)) {
 		/*
 		 * Can't use sync call while an async call is in flight
@@ -1048,8 +1046,15 @@ NTSTATUS cli_qpathinfo2(struct cli_state *cli, const char *fname,
 	if (!tevent_req_poll_ntstatus(req, ev, &status)) {
 		goto fail;
 	}
-	status = cli_qpathinfo2_recv(req, create_time, access_time,
-				     write_time, change_time, size, pattr, ino);
+	status = cli_qpathinfo2_recv(req,
+				     create_time,
+				     access_time,
+				     write_time,
+				     change_time,
+				     size,
+				     pattr,
+				     ino,
+				     mode);
  fail:
 	TALLOC_FREE(frame);
 	return status;
@@ -1065,6 +1070,7 @@ struct cli_qpathinfo_streams_state {
 };
 
 static void cli_qpathinfo_streams_done(struct tevent_req *subreq);
+static void cli_qpathinfo_streams_done2(struct tevent_req *subreq);
 
 struct tevent_req *cli_qpathinfo_streams_send(TALLOC_CTX *mem_ctx,
 					      struct tevent_context *ev,
@@ -1079,6 +1085,22 @@ struct tevent_req *cli_qpathinfo_streams_send(TALLOC_CTX *mem_ctx,
 	if (req == NULL) {
 		return NULL;
 	}
+	if (smbXcli_conn_protocol(cli->conn) >= PROTOCOL_SMB2_02) {
+		subreq = cli_smb2_qpathinfo_send(state,
+						 ev,
+						 cli,
+						 fname,
+						 FSCC_FILE_STREAM_INFORMATION,
+						 0,
+						 CLI_BUFFER_SIZE);
+		if (tevent_req_nomem(subreq, req)) {
+			return tevent_req_post(req, ev);
+		}
+		tevent_req_set_callback(subreq,
+					cli_qpathinfo_streams_done2,
+					req);
+		return req;
+	}
 	subreq = cli_qpathinfo_send(state, ev, cli, fname,
 				    SMB_FILE_STREAM_INFORMATION,
 				    0, CLI_BUFFER_SIZE);
@@ -1099,12 +1121,22 @@ static void cli_qpathinfo_streams_done(struct tevent_req *subreq)
 
 	status = cli_qpathinfo_recv(subreq, state, &state->data,
 				    &state->num_data);
-	TALLOC_FREE(subreq);
-	if (!NT_STATUS_IS_OK(status)) {
-		tevent_req_nterror(req, status);
-		return;
-	}
-	tevent_req_done(req);
+	tevent_req_simple_finish_ntstatus(subreq, status);
+}
+
+static void cli_qpathinfo_streams_done2(struct tevent_req *subreq)
+{
+	struct tevent_req *req =
+		tevent_req_callback_data(subreq, struct tevent_req);
+	struct cli_qpathinfo_streams_state *state =
+		tevent_req_data(req, struct cli_qpathinfo_streams_state);
+	NTSTATUS status;
+
+	status = cli_smb2_qpathinfo_recv(subreq,
+					 state,
+					 &state->data,
+					 &state->num_data);
+	tevent_req_simple_finish_ntstatus(subreq, status);
 }
 
 NTSTATUS cli_qpathinfo_streams_recv(struct tevent_req *req,
@@ -1136,14 +1168,6 @@ NTSTATUS cli_qpathinfo_streams(struct cli_state *cli, const char *fname,
 	struct tevent_req *req;
 	NTSTATUS status = NT_STATUS_NO_MEMORY;
 
-	if (smbXcli_conn_protocol(cli->conn) >= PROTOCOL_SMB2_02) {
-		return cli_smb2_qpathinfo_streams(cli,
-					fname,
-					mem_ctx,
-					pnum_streams,
-					pstreams);
-	}
-
 	frame = talloc_stackframe();
 
 	if (smbXcli_conn_has_async_calls(cli->conn)) {
@@ -1269,7 +1293,6 @@ struct cli_qfileinfo_basic_state {
 
 static void cli_qfileinfo_basic_done(struct tevent_req *subreq);
 static void cli_qfileinfo_basic_doneE(struct tevent_req *subreq);
-static void cli_qfileinfo_basic_done2(struct tevent_req *subreq);
 
 struct tevent_req *cli_qfileinfo_basic_send(
 	TALLOC_CTX *mem_ctx,
@@ -1311,34 +1334,13 @@ struct tevent_req *cli_qfileinfo_basic_send(
 		return req;
 	}
 
-	if (smbXcli_conn_protocol(cli->conn) >= PROTOCOL_SMB2_02) {
-		subreq = cli_smb2_query_info_fnum_send(
-			state,	/* mem_ctx */
-			ev,	/* ev */
-			cli,	/* cli */
-			fnum,	/* fnum */
-			1,	/* in_info_type */
-			(SMB_FILE_ALL_INFORMATION - 1000), /* in_file_info_class */
-			0xFFFF, /* in_max_output_length */
-			NULL,	/* in_input_buffer */
-			0,	/* in_additional_info */
-			0);	/* in_flags */
-		if (tevent_req_nomem(subreq, req)) {
-			return tevent_req_post(req, ev);
-		}
-		tevent_req_set_callback(
-			subreq, cli_qfileinfo_basic_done2, req);
-		return req;
-	}
-
-	subreq = cli_qfileinfo_send(
-		state,
-		ev,
-		cli,
-		fnum,
-		SMB_QUERY_FILE_ALL_INFO, /* level */
-		68,			 /* min_rdata */
-		CLI_BUFFER_SIZE);	 /* max_rdata */
+	subreq = cli_qfileinfo_send(state,
+				    ev,
+				    cli,
+				    fnum,
+				    FSCC_FILE_ALL_INFORMATION, /* level */
+				    68,			       /* min_rdata */
+				    CLI_BUFFER_SIZE);	       /* max_rdata */
 	if (tevent_req_nomem(subreq, req)) {
 		return tevent_req_post(req, ev);
 	}
@@ -1363,10 +1365,10 @@ static void cli_qfileinfo_basic_done(struct tevent_req *subreq)
 		return;
 	}
 
-	state->create_time = interpret_long_date((char *)rdata+0);
-	state->access_time = interpret_long_date((char *)rdata+8);
-	state->write_time = interpret_long_date((char *)rdata+16);
-	state->change_time = interpret_long_date((char *)rdata+24);
+	state->create_time = interpret_long_date(BVAL(rdata, 0));
+	state->access_time = interpret_long_date(BVAL(rdata, 8));
+	state->write_time = interpret_long_date(BVAL(rdata, 16));
+	state->change_time = interpret_long_date(BVAL(rdata, 24));
 	state->attr = PULL_LE_U32(rdata, 32);
 	state->size = PULL_LE_U64(rdata,48);
 	state->ino = PULL_LE_U32(rdata, 64);
@@ -1397,44 +1399,6 @@ static void cli_qfileinfo_basic_doneE(struct tevent_req *subreq)
 	tevent_req_done(req);
 }
 
-static void cli_qfileinfo_basic_done2(struct tevent_req *subreq)
-{
-	struct tevent_req *req = tevent_req_callback_data(
-		subreq, struct tevent_req);
-	struct cli_qfileinfo_basic_state *state = tevent_req_data(
-		req, struct cli_qfileinfo_basic_state);
-	DATA_BLOB outbuf = {0};
-	NTSTATUS status;
-
-	status = cli_smb2_query_info_fnum_recv(subreq, state, &outbuf);
-	TALLOC_FREE(subreq);
-	if (tevent_req_nterror(req, status)) {
-		return;
-	}
-
-	/* Parse the reply. */
-	if (outbuf.length < 0x60) {
-		tevent_req_nterror(req, NT_STATUS_INVALID_NETWORK_RESPONSE);
-		return;
-	}
-
-	state->create_time = interpret_long_date(
-		(const char *)outbuf.data + 0x0);
-	state->access_time = interpret_long_date(
-		(const char *)outbuf.data + 0x8);
-	state->write_time = interpret_long_date(
-		(const char *)outbuf.data + 0x10);
-	state->change_time = interpret_long_date(
-		(const char *)outbuf.data + 0x18);
-	state->attr = IVAL(outbuf.data, 0x20);
-	state->size = BVAL(outbuf.data, 0x30);
-	state->ino = BVAL(outbuf.data, 0x40);
-
-	data_blob_free(&outbuf);
-
-	tevent_req_done(req);
-}
-
 NTSTATUS cli_qfileinfo_basic_recv(
 	struct tevent_req *req,
 	uint32_t *attr,
@@ -1527,9 +1491,6 @@ NTSTATUS cli_qfileinfo_basic(
 		write_time,
 		change_time,
 		ino);
-
-	/* cli_smb2_query_info_fnum_recv doesn't set this */
-	cli->raw_status = status;
 fail:
 	TALLOC_FREE(frame);
 	return status;
@@ -1580,8 +1541,7 @@ static void cli_qpathinfo_basic_done(struct tevent_req *subreq)
 	status = cli_qpathinfo_recv(subreq, state, &state->data,
 				    &state->num_data);
 	TALLOC_FREE(subreq);
-	if (!NT_STATUS_IS_OK(status)) {
-		tevent_req_nterror(req, status);
+	if (tevent_req_nterror(req, status)) {
 		return;
 	}
 	tevent_req_done(req);
@@ -1598,10 +1558,10 @@ NTSTATUS cli_qpathinfo_basic_recv(struct tevent_req *req,
 		return status;
 	}
 
-	sbuf->st_ex_btime = interpret_long_date((char *)state->data);
-	sbuf->st_ex_atime = interpret_long_date((char *)state->data+8);
-	sbuf->st_ex_mtime = interpret_long_date((char *)state->data+16);
-	sbuf->st_ex_ctime = interpret_long_date((char *)state->data+24);
+	sbuf->st_ex_btime = interpret_long_date(BVAL(state->data, 0));
+	sbuf->st_ex_atime = interpret_long_date(BVAL(state->data, 8));
+	sbuf->st_ex_mtime = interpret_long_date(BVAL(state->data, 16));
+	sbuf->st_ex_ctime = interpret_long_date(BVAL(state->data, 24));
 	*attributes = IVAL(state->data, 32);
 	return NT_STATUS_OK;
 }
@@ -1660,12 +1620,6 @@ NTSTATUS cli_qpathinfo_alt_name(struct cli_state *cli, const char *fname, fstrin
 	size_t converted_size = 0;
 	NTSTATUS status;
 
-	if (smbXcli_conn_protocol(cli->conn) >= PROTOCOL_SMB2_02) {
-		return cli_smb2_qpathinfo_alt_name(cli,
-						fname,
-						alt_name);
-	}
-
 	status = cli_qpathinfo(talloc_tos(), cli, fname,
 			       SMB_QUERY_FILE_ALT_NAME_INFO,
 			       4, CLI_BUFFER_SIZE, &rdata, &num_rdata);
@@ -1701,10 +1655,13 @@ NTSTATUS cli_qpathinfo_alt_name(struct cli_state *cli, const char *fname, fstrin
  Send a qpathinfo SMB_QUERY_FILE_STANDARD_INFO call.
 ****************************************************************************/
 
-NTSTATUS cli_qpathinfo_standard(struct cli_state *cli, const char *fname,
-				uint64_t *allocated, uint64_t *size,
-				uint32_t *nlinks,
-				bool *is_del_pending, bool *is_dir)
+static NTSTATUS cli_qpathinfo_standard(struct cli_state *cli,
+				       const char *fname,
+				       uint64_t *allocated,
+				       uint64_t *size,
+				       uint32_t *nlinks,
+				       bool *is_del_pending,
+				       bool *is_dir)
 {
 	uint8_t *rdata;
 	uint32_t num_rdata;
@@ -1768,9 +1725,16 @@ NTSTATUS cli_qpathinfo3(struct cli_state *cli, const char *fname,
 		 * an intermediate attr variable as below but can
 		 * pass pattr directly.
 		 */
-		return cli_qpathinfo2(cli, fname,
-				      create_time, access_time, write_time, change_time,
-				      size, pattr, ino);
+		return cli_qpathinfo2(cli,
+				      fname,
+				      create_time,
+				      access_time,
+				      write_time,
+				      change_time,
+				      size,
+				      pattr,
+				      ino,
+				      NULL);
 	}
 
 	if (create_time || access_time || write_time || change_time || pattr) {
