@@ -416,7 +416,7 @@ static int streams_xattr_openat(struct vfs_handle_struct *handle,
 					   how);
 	}
 
-	if (how->resolve != 0) {
+	if ((how->resolve & ~VFS_OPEN_HOW_WITH_BACKUP_INTENT) != 0) {
 		errno = ENOSYS;
 		return -1;
 	}
