@@ -87,13 +87,18 @@ ssize_t pwrite_fsync_recv(struct tevent_req *req, int *perr);
 
 /* The following definitions come from smbd/blocking.c  */
 
-NTSTATUS smbd_do_locks_try(
-	struct files_struct *fsp,
-	uint16_t num_locks,
-	struct smbd_lock_element *locks,
-	uint16_t *blocker_idx,
-	struct server_id *blocking_pid,
-	uint64_t *blocking_smblctx);
+struct smbd_do_locks_state {
+	uint16_t num_locks;
+	struct smbd_lock_element *locks;
+	NTSTATUS status;
+	uint16_t blocker_idx;
+	struct server_id blocking_pid;
+	uint64_t blocking_smblctx;
+};
+
+NTSTATUS smbd_do_locks_try(struct byte_range_lock *br_lck,
+			   struct smbd_do_locks_state *state);
+
 struct tevent_req *smbd_smb1_do_locks_send(
 	TALLOC_CTX *mem_ctx,
 	struct tevent_context *ev,
@@ -134,12 +139,14 @@ void msg_close_file(struct messaging_context *msg_ctx,
 NTSTATUS delete_all_streams(struct files_struct *fsp,
 			    struct files_struct *dirfsp,
 			    struct smb_filename *fsp_atname);
-NTSTATUS recursive_rmdir(TALLOC_CTX *ctx,
-		     connection_struct *conn,
-		     struct smb_filename *smb_dname);
+NTSTATUS recursive_rmdir_fsp(struct files_struct *fsp);
 bool has_other_nonposix_opens(struct share_mode_lock *lck,
 			      struct files_struct *fsp);
 bool has_nonposix_opens(struct share_mode_lock *lck);
+bool has_delete_opens(struct files_struct *fsp,
+		      struct share_mode_lock *lck,
+		      bool *delete_access,
+		      bool *delete_pending);
 
 /* The following definitions come from smbd/conn.c  */
 
@@ -222,7 +229,7 @@ NTSTATUS file_set_sparse(connection_struct *conn,
 int file_ntimes(connection_struct *conn,
 		files_struct *fsp,
 		struct smb_file_time *ft);
-bool set_sticky_write_time_fsp(struct files_struct *fsp,
+void set_sticky_write_time_fsp(struct files_struct *fsp,
 			       struct timespec mtime);
 
 NTSTATUS fget_ea_dos_attribute(struct files_struct *fsp,
@@ -238,10 +245,6 @@ struct timespec get_create_timespec(connection_struct *conn,
 				struct files_struct *fsp,
 				const struct smb_filename *smb_fname);
 
-struct timespec get_change_timespec(connection_struct *conn,
-				struct files_struct *fsp,
-				const struct smb_filename *smb_fname);
-
 NTSTATUS parse_dos_attribute_blob(struct smb_filename *smb_fname,
 				  DATA_BLOB blob,
 				  uint32_t *pattr);
@@ -276,10 +279,15 @@ NTSTATUS can_set_delete_on_close(files_struct *fsp, uint32_t dosmode);
 /* The following definitions come from smbd/fileio.c  */
 
 ssize_t read_file(files_struct *fsp,char *data,off_t pos,size_t n);
-void fsp_flush_write_time_update(struct files_struct *fsp);
-void trigger_write_time_update(struct files_struct *fsp);
-void trigger_write_time_update_immediate(struct files_struct *fsp);
-void mark_file_modified(files_struct *fsp);
+void trigger_write_time_update_immediate(struct files_struct *fsp,
+					 bool update_mtime,
+					 bool update_ctime);
+struct file_modified_state;
+void prepare_file_modified(files_struct *fsp,
+			   struct file_modified_state *state);
+void mark_file_modified(files_struct *fsp,
+			bool modified,
+			struct file_modified_state *state);
 ssize_t write_file(struct smb_request *req,
 			files_struct *fsp,
 			const char *data,
@@ -393,6 +401,10 @@ NTSTATUS openat_pathref_fsp_nosymlink(
 NTSTATUS openat_pathref_fsp_lcomp(struct files_struct *dirfsp,
 				  struct smb_filename *smb_fname_rel,
 				  uint32_t ucf_flags);
+NTSTATUS openat_pathref_fsp_dot(TALLOC_CTX *mem_ctx,
+				struct files_struct *dirfsp,
+				uint32_t flags,
+				struct smb_filename **_dot);
 NTSTATUS readlink_talloc(
 	TALLOC_CTX *mem_ctx,
 	struct files_struct *dirfsp,
@@ -427,6 +439,9 @@ NTSTATUS parent_pathref(TALLOC_CTX *mem_ctx,
 			struct smb_filename **_parent,
 			struct smb_filename **_atname);
 
+void fsp_apply_private_ntcreatex_flags(struct files_struct *fsp,
+				       uint32_t flags);
+
 /* The following definitions come from smbd/smb2_ipc.c  */
 
 NTSTATUS nt_status_np_pipe(NTSTATUS status);
@@ -862,7 +877,8 @@ void process_smb(struct smbXsrv_connection *xconn,
 void smbd_process(struct tevent_context *ev_ctx,
 		  struct messaging_context *msg_ctx,
 		  int sock_fd,
-		  bool interactive);
+		  bool interactive,
+		  enum smb_transport_type transport_type);
 bool valid_smb1_header(const uint8_t *inbuf);
 bool init_smb1_request(struct smb_request *req,
 		      struct smbd_server_connection *sconn,
@@ -1238,6 +1254,7 @@ struct tevent_req *delay_for_handle_lease_break_send(
 	struct tevent_context *ev,
 	struct timeval timeout,
 	struct files_struct *fsp,
+	uint32_t access_mask,
 	bool recursive,
 	struct share_mode_lock **lck);
 
